unit DataSetReader;

interface

uses
  System.TypInfo, System.Generics.Collections, Data.DB;

type
  DBFieldAttribute = class(TCustomAttribute)
  private
    FName: string;
    FLoad: Boolean;
  public
    constructor Create(const AName: string); overload;
    constructor Create(const ALoad: Boolean); overload;
    property Load: Boolean read FLoad;
    property Name: string read FName;
  end;

  TRecordDataSetReaderBase = class(TObject)
  public type
    TOption = (IgnoreMissingFields);
    TOptions = set of TOption;
  private type
    TFieldInfo = class(TObject)
      FOffset: Cardinal;
      FIndex: Integer;
      FType: TFieldType;
    end;
  private
    FTypeInfo: PTypeInfo;
    FInfos: TList<TFieldInfo>;
  protected
    constructor Create(ATypeInfo: PTypeInfo);
    procedure Bind(ADataSet: TDataSet; AOptions: TOptions);
    procedure Read(ADataSet: TDataSet; ApRec: Pointer; AOptions: TOptions); overload;
  public
    destructor Destroy; override;
  end;

  TRecordDataSetReader<T> = class(TRecordDataSetReaderBase)
  public
    constructor Create;
    procedure Read(ADataSet: TDataSet; var ARec: T;
      AOptions: TRecordDataSetReaderBase.TOptions = []); overload; inline;
  end;

implementation

uses
  System.Variants, System.SysUtils, System.Classes, System.Rtti;

{ DBFieldAttribute }

constructor DBFieldAttribute.Create(const AName: string);
begin
  inherited Create;
  FName := AName;
  FLoad := True;
end;

constructor DBFieldAttribute.Create(const ALoad: Boolean);
begin
  inherited Create;
  FLoad := ALoad;
end;

{ TRecordDataSetReaderBase }

constructor TRecordDataSetReaderBase.Create(ATypeInfo: PTypeInfo);
begin
  inherited Create;
  FTypeInfo := ATypeInfo;
end;

destructor TRecordDataSetReaderBase.Destroy;
begin
  FreeAndNil(FInfos);
  inherited Destroy;
end;

procedure TRecordDataSetReaderBase.Bind(ADataSet: TDataSet; AOptions: TOptions);
var
  LCtx: TRttiContext;
  LType: TRttiType;
  LAttr: TCustomAttribute;
  LField: TRttiField;
  LDSField: TField;
  LDSType: TFieldType;
  LInfo: TFieldInfo;
  LDSPrefix: string;
  LDSName: string;
  LLoad: Boolean;
  LErrors: TStringList;
begin
  LErrors := TStringList.Create;
  try
    LCtx := TRttiContext.Create;
    LType := LCtx.GetType(FTypeInfo);
    LDSPrefix := '';
    for LAttr in LType.GetAttributes do
      if LAttr is DBFieldAttribute then
      begin
        LDSPrefix := DBFieldAttribute(LAttr).Name;
        if LDSPrefix.EndsWith('*') then
          LDSPrefix := LDSPrefix.Substring(0, LDSPrefix.Length - 1);
        Break;
      end;

    for LField in LType.GetFields do
    begin
      LDSName := LDSPrefix + LField.Name;
      LLoad := True;
      for LAttr in LField.GetAttributes do
        if LAttr is DBFieldAttribute then
        begin
          LDSName := DBFieldAttribute(LAttr).Name;
          LLoad := DBFieldAttribute(LAttr).Load;
          Break;
        end;
      if (LDSName = '') or not LLoad then
        Continue;

      LDSField := ADataSet.FindField(LDSName);
      if LDSField = nil then
      begin
        if not (TOption.IgnoreMissingFields in AOptions) then
          LErrors.Add(Format('"%s.%s" field is not mapped to a DB field',
            [LType.Name, LField.Name]));
        Continue;
      end;

      case LField.FieldType.TypeKind of
      tkInteger:
        case GetTypeData(LField.FieldType.Handle)^.OrdType of
          otSByte: LDSType := ftShortint;
          otUByte: LDSType := ftByte;
          otSWord: LDSType := ftSmallint;
          otUWord: LDSType := ftWord;
          otSLong: LDSType := ftInteger;
          otULong: LDSType := ftLongWord;
        else
          LErrors.Add(Format('tkInteger data subtype of "%s.%s" field is not supported',
            [LType.Name, LField.Name]));
          Continue;
        end;
      tkFloat:
        case GetTypeData(LField.FieldType.Handle)^.FloatType of
          TFloatType.ftSingle: LDSType := TFieldType.ftSingle;
          TFloatType.ftDouble: LDSType := TFieldType.ftFloat;
          TFloatType.ftExtended: LDSType := TFieldType.ftExtended;
          TFloatType.ftCurr: LDSType := TFieldType.ftCurrency;
        else
          LErrors.Add(Format('tkFloat data subtype of "%s.%s" field is not supported',
            [LType.Name, LField.Name]));
          Continue;
        end;
      tkUString: LDSType := ftString;
      tkInt64: LDSType := ftLargeint;
      tkEnumeration:
        if LField.FieldType.Handle = System.TypeInfo(Boolean) then
          LDSType := ftBoolean
        else
        begin
          LErrors.Add(Format('tkEnumeration data subtype of "%s.%s" field is not supported',
            [LType.Name, LField.Name]));
          Continue;
        end;
      else
        LErrors.Add(Format(GetEnumName(TypeInfo(System.TTypeKind), Integer(LField.FieldType.TypeKind)) +
          ' data type of "%s.%s" field is not supported', [LType.Name, LField.Name]));
        Continue;
      end;

      if LDSType <> ftUnknown then
      begin
        if FInfos = nil then
          FInfos := TObjectList<TFieldInfo>.Create(True);
        LInfo := TFieldInfo.Create;
        LInfo.FOffset := LField.Offset;
        LInfo.FIndex := LDSField.Index;
        LInfo.FType := LDSType;
        FInfos.Add(LInfo);
      end;
    end;
    if FInfos = nil then
      LErrors.Add(Format('No dataset fields matching to record "%s" fields',
        [LType.Name]));
    if LErrors.Count > 0 then
      raise EDatabaseError.Create(LErrors.Text);
  finally
    LErrors.Free;
  end;
end;

procedure TRecordDataSetReaderBase.Read(ADataSet: TDataSet; ApRec: Pointer; AOptions: TOptions);
var
  i: Integer;
  LInfo: TFieldInfo;
  LFields: TFields;
  LField: TField;
begin
  if FInfos = nil then
    Bind(ADataSet, AOptions);
  LFields := ADataSet.Fields;
  for i := 0 to FInfos.Count - 1 do
  begin
    LInfo := FInfos.List[i];
    LField := LFields[LInfo.FIndex];
    case LInfo.FType of
      ftString:   PString(PByte(ApRec) + LInfo.FOffset)^ := LField.AsString;
      ftShortint: PShortInt(PByte(ApRec) + LInfo.FOffset)^ := LField.AsInteger;
      ftByte:     PByte(PByte(ApRec) + LInfo.FOffset)^ := LField.AsLongWord;
      ftSmallint: PSmallInt(PByte(ApRec) + LInfo.FOffset)^ := LField.AsInteger;
      ftWord:     PWord(PByte(ApRec) + LInfo.FOffset)^ := LField.AsLongWord;
      ftInteger:  PInteger(PByte(ApRec) + LInfo.FOffset)^ := LField.AsInteger;
      ftLongWord: PCardinal(PByte(ApRec) + LInfo.FOffset)^ := LField.AsLongWord;
      ftSingle:   PSingle(PByte(ApRec) + LInfo.FOffset)^ := LField.AsSingle;
      ftFloat:    PDouble(PByte(ApRec) + LInfo.FOffset)^ := LField.AsFloat;
      ftExtended: PExtended(PByte(ApRec) + LInfo.FOffset)^ := LField.AsExtended;
      ftCurrency: PCurrency(PByte(ApRec) + LInfo.FOffset)^ := LField.AsCurrency;
      ftLargeint: PInt64(PByte(ApRec) + LInfo.FOffset)^ := LField.AsLargeInt;
      ftBoolean:  PBoolean(PByte(ApRec) + LInfo.FOffset)^ := LField.AsBoolean;
      else        Assert(False);
    end;
  end;
end;

{ TRecordDataSetReader }

constructor TRecordDataSetReader<T>.Create;
begin
  inherited Create(TypeInfo(T));
end;

procedure TRecordDataSetReader<T>.Read(ADataSet: TDataSet; var ARec: T;
  AOptions: TRecordDataSetReaderBase.TOptions = []);
begin
  inherited Read(ADataSet, @ARec, AOptions);
end;

end.

